#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright 1999, 2000 Martin Pool <mbp@humbug.org.au>
# Copyright 2002 Gerardo Poggiali
# Copyright 2007, 2008, 2009 Bernie Innocenti <bernie@codewiz.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__version__ = '$Id: 678754e3a9633d29fc779eeed48eeddb8fa9fa87 $'[4:12]

from time import clock
start_time = clock()
title_done = False

import cgi, sys, os, re, errno, stat

# FIXME: we accept stuff like foo/../bar and we shouldn't
file_re = re.compile(r"^\b([A-Za-z0-9_\-][A-Za-z0-9_\.\-/]*)\b$")
img_re = re.compile(r"^.*\.(png|gif|jpg|jpeg|bmp|ico|ogm|ogg|mkv|mpg|mpeg|mp4|avi|asf|flv|wmv|qt)$", re.IGNORECASE)
video_re = re.compile(r"^.*\.(ogm|ogg|mkv|mpg|mpeg|mp4|avi|asf|flv|wmv|qt)$", re.IGNORECASE)
url_re = re.compile(r"^[a-z]{3,8}://[^\s'\"]+\S$")
ext_re = re.compile(r"\.([^\./]+)$")

# CGI stuff ---------------------------------------------------------
def script_name():
    return os.environ.get('SCRIPT_NAME', '')

def query_string():
    path_info = os.environ.get('PATH_INFO', '')
    if len(path_info) and path_info[0] == '/':
        return path_info[1:] or 'FrontPage'
    else:
        return os.environ.get('QUERY_STRING', '') or 'FrontPage'

def privileged_path():
    return privileged_url or script_name()

def remote_user():
    user = os.environ.get('REMOTE_USER', '')
    if user is None or user == '' or user == 'anonymous':
        user = 'AnonymousCoward'
    return user

def remote_host():
    return os.environ.get('REMOTE_ADDR', '')

def get_hostname(addr):
    try:
        from socket import gethostbyaddr
        return gethostbyaddr(addr)[0] + ' (' + addr + ')'
    except Exception:
        return addr

def is_external_url(pathname):
    return (url_re.match(pathname) or pathname.startswith('/'))

def relative_url(pathname, privileged=False):
    if not is_external_url(pathname):
        if privileged:
            url = privileged_path()
        else:
            url = script_name()
        pathname = url + '/' + pathname
    return cgi.escape(pathname, quote=True)

def permalink(s):
    return re.sub(' ', '-', re.sub('[^a-z0-9_ ]', '', s.lower()).strip())

def humanlink(s):
    return re.sub(r'([^:/\.]+)(?:\.[^/:]+|)$', r'\1', s.replace('_', ' '))

# Split arg lists like "blah| blah blah| width=100 | align = center",
# return a list containing anonymous arguments and a map containing the named arguments
def parse_args(s):
    args = []
    kwargs = {} 
    for arg in s.strip('<[{}]>').split('|'):
        m = re.match('\s*(\w+)\s*=\s*(.+)\s*', arg)
        if m is not None:
            kwargs[m.group(1)] = m.group(2)
        else:
            args.append(arg.strip())
    return (args, kwargs)

def url_args(kvargs):
    argv = []
    for k, v in kvargs.items():
        argv.append(k + '=' + v)
    if argv:
        return '?' + '&amp;'.join(argv)
    return ''

# Formatting stuff --------------------------------------------------
def emit_header(mime_type="text/html"):
    print("Content-type: " + mime_type + "; charset=utf-8\n")

def send_guru(msg_text, msg_type):
    if not msg_text: return
    print('<pre id="guru" onclick="this.style.display = \'none\'" class="' + msg_type + '">')
    if msg_type == 'error':
        print('    Software Failure.  Press left mouse button to continue.\n')
    print(msg_text)
    if msg_type == 'error':
        print '\n           Guru Meditation #DEADBEEF.ABADC0DE'
    print('</pre><script language="JavaScript" type="text/javascript" src="%s" defer="defer"></script>' \
        % relative_url('sys/GuruMeditation.js'))

def send_title(name, text="Limbo", msg_text=None, msg_type='error', writable=False):
    global title_done
    if title_done: return

    # Head
    emit_header()
    print('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"')
    print('  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">')
    print('<html xmlns="http://www.w3.org/1999/xhtml" lang="en" xml:lang="en">')

    print("<head><title>%s: %s</title>" % (site_name, text))
    print(' <meta http-equiv="Content-Type" content="application/xhtml+xml; charset=utf-8" />')
    if not name:
        print(' <meta name="robots" content="noindex,nofollow" />')

    for meta in meta_urls:
        http_equiv, content = meta
        print(' <meta http-equiv="%s" content="%s" />' % (http_equiv, relative_url(content)))

    for link in link_urls:
        rel, href = link
        print(' <link rel="%s" href="%s" />' % (rel, relative_url(href)))

    if name and writable and privileged_url is not None:
        print(' <link rel="alternate" type="application/x-wiki" title="Edit this page" href="%s" />' \
            % (privileged_path() + '?a=edit&q=' + name))

    if history_url is not None:
        print(' <link rel="alternate" type="application/rss+xml" title="RSS" href="%s" />' \
            % relative_url(history_url + '?a=rss'))

    print('</head>')

    # Body
    if name and writable and privileged_url is not None:
        print('<body ondblclick="location.href=\'' + privileged_path() + '?a=edit&q=' + name + '\'">')
    else:
        print('<body>')

    title_done = True
    send_guru(msg_text, msg_type)

    # Navbar
    print('<div class="nav">')
    print link_tag('FrontPage', site_icon or 'Home', cssclass='navlink')
    if name:
        print('  <b>' + link_tag('?fullsearch=' + name, text, cssclass='navlink') + '</b> ')
    else:
        print('  <b>' + text + '</b> ')
    print(' | ' + link_tag('FindPage', 'Find Page', cssclass='navlink'))
    if 'history_url' in globals():
        print(' | <a href="' + relative_url(history_url) + '" class="navlink">Recent Changes</a>')
        if name:
            print(' | <a href="' + relative_url(history_url + '?a=history;f=' + name) + '" class="navlink">Page History</a>')

    if name:
        print(' | ' + link_tag(name + '?a=raw', 'Raw Text', cssclass='navlink'))
        if privileged_url is not None:
            if writable:
                print(' | ' + link_tag('?a=edit&q=' + name, 'Edit', cssclass='navlink', privileged=True))
            else:
                print(' | ' + link_tag(name, 'Login', cssclass='navlink', privileged=True))

    else:
        print(' | <i>Immutable Page</i>')

    user = remote_user()
    if user != 'AnonymousCoward':
        print(' | <span class="login"><i><b>' + link_tag('User/' + user, user) + '</b></i></span>')

    print('<hr /></div>')

def send_httperror(status="403 Not Found", query=""):
    print("Status: %s" % status)
    send_title(None, msg_text=("%s: on query '%s'" % (status, query)))
    send_footer()

def link_tag(dest, text=None, privileged=False, **kvargs):
    if text is None:
        text = humanlink(dest)
    elif img_re.match(text):
        text = '<img border="0" src="' + relative_url(text) + '" alt="' + text + '" />'

    link_class = kvargs.get('class', kvargs.get('cssclass', None))
    if not link_class:
        if is_external_url(dest):
            link_class = 'external'
        elif file_re.match(dest) and Page(dest).exists():
            link_class = 'wikilink'
        else:
            text = nonexist_pfx + text
            link_class = 'nonexistent'

    # Prevent crawlers from following links potentially added by spammers or to generated pages
    nofollow = ''
    if link_class == 'external' or link_class == 'navlink':
        nofollow = 'rel="nofollow" '

    return '<a class="%s" %shref="%s">%s</a>' % (link_class, nofollow, relative_url(dest, privileged=privileged), text)

def link_inline(name, descr=None, kvargs={}):
    if not descr: descr = humanlink(name)
    url = relative_url(name)
    if video_re.match(name):
        return '<video controls="1" src="%s">Your browser does not support the HTML5 video tag</video>' % url
    elif img_re.match(name):
        return '<a href="%s"><img border="0" src="%s" alt="%s" /></a>' % (url, url + url_args(kvargs), descr)
    elif file_re.match(name) and not ext_re.search(name): # FIXME: this guesses a wiki page
        return Page(name).send_naked()
    else:
        return '<iframe width="100%%" scrolling="auto" frameborder="0" src="%s"><a href="%s">%s</a></iframe>' \
            % (url, url, name)

# Search ---------------------------------------------------

def print_search_stats(hits, searched):
    print("<p>%d hits out of %d pages searched.</p>" % (hits, searched))

def handle_fullsearch(query, form):
    needle = form['q'].value
    send_title(None, 'Full text search for "' + needle + '"')

    needle_re = re.compile(needle, re.IGNORECASE)
    hits = []
    all_pages = page_list()
    for page_name in all_pages:
        body = Page(page_name).get_raw_body()
        count = len(needle_re.findall(body))
        if count:
            hits.append((count, page_name))

    # The default comparison for tuples compares elements in order,
    # so this sorts by number of hits
    hits.sort()
    hits.reverse()

    print("<ul>")
    for (count, page_name) in hits:
        print('<li><p>' + link_tag(page_name))
        print(' . . . . ' + `count`)
        print(['match', 'matches'][count != 1])
        print('</p></li>')
    print("</ul>")

    print_search_stats(len(hits), len(all_pages))

def handle_titlesearch(query, form):
    needle = form['q'].value
    send_title(None, 'Title search for "' + needle + '"')

    needle_re = re.compile(needle, re.IGNORECASE)
    all_pages = page_list()
    hits = list(filter(needle_re.search, all_pages))

    print("<ul>")
    for filename in hits:
        print('<li><p>' + link_tag(filename) + "</p></li>")
    print("</ul>")

    print_search_stats(len(hits), len(all_pages))

def handle_raw(pagename, form):
    if not file_re.match(pagename):
        send_httperror("403 Forbidden", pagename)
        return

    Page(pagename).send_raw()

def handle_edit(pagename, form):
    if not file_re.match(pagename):
        send_httperror("403 Forbidden", pagename)
        return

    pg = Page(form['q'].value)
    if 'save' in form:
        if form['file'].value:
            pg.save(form['file'].file.read(), form['changelog'].value)
        else:
            pg.save(form['savetext'].value.replace('\r\n', '\n'), form['changelog'].value)
        pg.format()
    elif 'cancel' in form:
        pg.msg_text = 'Editing canceled'
        pg.msg_type = 'notice'
        pg.format()
    else: # preview or edit
        text = None
        if 'preview' in form:
            text = form['savetext'].value
        pg.send_editor(text)

def handle_get(pagename, form):
        if file_re.match(pagename):
            # FIMXE: this is all bullshit, MimeTypes bases its guess on the extension!
            from mimetypes import MimeTypes
            mimetype, encoding = MimeTypes().guess_type(pagename)
            if mimetype:
                Page(pagename).send_raw(mimetype=mimetype, args=form)
            else:
                Page(pagename).format()
        else:
            send_httperror("403 Forbidden", pagename)

# Used by macros/WordIndex and macros/TitleIndex
def make_index_key():
    links = ['<a href="#%s">%s</a>' % (ch, ch) for ch in 'abcdefghijklmnopqrstuvwxyz']
    return '<p style="text-align: center">' + ' | '.join(links) + '</p>'

def page_list(dirname = None, re = None):
    if re is None:
        # FIXME: WikiWord is too restrictive now!
        re = re.compile(r"^\b((([A-Z][a-z0-9]+){2,}/)*([A-Z][a-z0-9]+){2,})\b$")
    return sorted(filter(re.match, os.listdir(dirname or data_dir)))

def send_footer(mod_string=None):
    if globals().get('debug_cgi', False):
        cgi.print_arguments()
        cgi.print_form(form)
        cgi.print_environ()
    print('''
<div id="footer"><hr />
<p class="copyright">
<a rel="license" href="http://creativecommons.org/licenses/by-sa/3.0/"><img class="license" alt="Creative Commons License" src="%s" /></a>
<span class="benchmark">generated in %0.3fs</span> by <a href="http://www.codewiz.org/wiki/GeekiGeeki">GeekiGeeki</a> version %s
</p>
''' % (relative_url('cc-by-sa.png'), clock() - start_time, __version__))
    if mod_string:
        print('<p class="modified">last modified %s</p>' % mod_string)
    print('</div></body></html>')

class WikiFormatter:
    """Object that turns Wiki markup into HTML.

    All formatting commands can be parsed one line at a time, though
    some state is carried over between lines.
    """
    def __init__(self, raw):
        self.raw = raw
        self.h_level = 0
        self.in_pre = self.in_html = self.in_table = self.in_li = False
        self.in_header = True
        self.list_indents = []
        self.tr_cnt = 0
        self.styles = {
            #wiki   html   enabled?
            "//":  ["em",  False],
            "**":  ["b",   False],
            "##":  ["tt",  False],
            "__":  ["u",   False],
            "^^":  ["sup", False],
            ",,":  ["sub", False],
            "''":  ["em",  False], # LEGACY
            "'''": ["b",   False], # LEGACY
            "``":  ["tt",  False], # LEGACY
        }

    def _b_repl(self, word):
        style = self.styles[word]
        style[1] = not style[1]
        return ['</', '<'][style[1]] + style[0] + '>'

    def _tit_repl(self, word):
        if self.h_level:
            result = '</h%d><p>\n' % self.h_level
            self.h_level = 0
        else:
            self.h_level = len(word) - 1
            link = permalink(self.line)
            result = '\n</p><h%d id="%s"><a class="heading" href="#%s">¶</a> ' % (self.h_level, link, link)
        return result

    def _br_repl(self, word):
        return '<br />'

    def _rule_repl(self, word):
        return self._undent() + '\n<hr size="%d" noshade="noshade" />\n' % (len(word) - 2)

    def _macro_repl(self, word):
        try:
            args, kwargs = parse_args(word)
            macro = globals().get('_macro_' + args[0])
            if not macro:
                exec(open("macros/" + name + ".py").read(), globals())
                macro = globals().get('_macro_' + name)
            return macro(*args, **kwargs)
        except Exception:
            msg = cgi.escape(word)
            if not self.in_html:
                msg = '<strong class="error">' + msg + '</strong>'
            return msg

    def _hurl_repl(self, word):
        args, kvargs = parse_args(word)
        return link_tag(*args, **kvargs)

    def _inl_repl(self, word):
        args, kvargs = parse_args(word)
        name = args.pop(0)
        if len(args):
            descr = args.pop(0)
            # This double div nonsense works around a limitation of the HTML block model
            return '<div class="' + kvargs.get('class', 'thumb') + '">' \
                + '<div class="innerthumb">' \
                + link_inline(name, descr, kvargs) \
                + '<div class="caption">' + descr + '</div></div></div>'
        else:
            return link_inline(name, None, kvargs)

    def _html_repl(self, word):
        if not self.in_html and word.startswith('<div'): word = '</p>' + word
        self.in_html += 1
        return word; # Pass through

    def _htmle_repl(self, word):
        self.in_html -= 1
        if not self.in_html and word.startswith('</div'): word += '<p>'
        return word; # Pass through

    def _ent_repl(self, s):
        if self.in_html:
            return s; # Pass through
        return {'&': '&amp;',
                '<': '&lt;',
                '>': '&gt;'}[s]

    def _img_repl(self, word): # LEGACY
        return self._inl_repl('{{' + word + '}}')

    def _word_repl(self, word): # LEGACY
        if self.in_html: return word # pass through
        return link_tag(word)

    def _url_repl(self, word): # LEGACY
        if self.in_html: return word # pass through
        return link_tag(word)

    def _email_repl(self, word): # LEGACY
        if self.in_html: return word # pass through
        return '<a href="mailto:%s">%s</a>' % (word, word)

    def _li_repl(self, match):
        if self.in_li:
            return '</li><li>'
        else:
            self.in_li = True
            return '<li>'

    def _pre_repl(self, word):
        if word == '{{{' and not self.in_pre:
            self.in_pre = True
            return '<pre>'
        elif self.in_pre:
            self.in_pre = False
            return '</pre>'
        return ''

    def _hi_repl(self, word):
        return '<strong class="highlight ' + word + '">' + word + '</strong>'

    def _tr_repl(self, word):
        out = ''
        if not self.in_table:
            self.in_table = True
            self.tr_cnt = 0
            out = '</p><table><tbody>\n'
        self.tr_cnt += 1
        out = out + '<tr class="' + ['even', 'odd'][self.tr_cnt % 2] + '">'
        return out + ['<td>', '<th>'][word.strip() == '||=']

    def _td_repl(self, word):
        if self.in_table:
            return ['</td><td>', '</th><th>'][word.strip() == '||=']
        return ''

    def _tre_repl(self, word):
        if self.in_table:
            return ['</td></tr>', '</th></tr>'][word.strip() == '||=']
        return ''

    def _indent_level(self):
        return len(self.list_indents) and self.list_indents[-1]

    def _indent_to(self, new_level):
        if self._indent_level() == new_level:
            return ''
        s = '</p>'
        while self._indent_level() > new_level:
            del(self.list_indents[-1])
            if self.in_li:
                s += '</li>'
                self.in_li = False # FIXME
            s += '</ul>\n'
        while self._indent_level() < new_level:
            self.list_indents.append(new_level)
            s += '<ul>\n'
        s += '<p>'
        return s

    def _undent(self):
        res = '</p>'
        res += '</ul>' * len(self.list_indents)
        res += '<p>'
        self.list_indents = []
        return res

    def replace(self, match):
        for rule, hit in list(match.groupdict().items()):
            if hit:
                return getattr(self, '_' + rule + '_repl')(hit)
        else:
            raise "Can't handle match " + repr(match)

    def print_html(self):
        print('<div class="wiki"><p>')

        scan_re = re.compile(r"""(?:
            # Styles and formatting
              (?P<b>     \*\*|'''|//|''|\#\#|``|__|\^\^|,,)
            | (?P<tit>   \={2,6})
            | (?P<br>    \\\\)
            | (?P<rule>  ^-{3,})
            | (?P<hi>    \b( FIXME | TODO | DONE )\b )

            # Links
            | (?P<macro> \<\<([^\s\|\>]+)(?:\s*\|\s*([^\>]+)|)\>\>)
            | (?P<hurl>  \[\[([^\s\|]+)(?:\s*\|\s*([^\]]+)|)\]\])

            # Inline HTML
            | (?P<html>  <(br|hr|div|span|form|iframe|input|textarea|a|img|h[1-5])\b )
            | (?P<htmle> ( /\s*> | </(br|hr|div|span|form|iframe|input|textarea|a|img|h[1-5])> ) )
            | (?P<ent>   [<>&] )

            # Auto links (LEGACY)
            | (?P<img>   \b[a-zA-Z0-9_/-]+\.(png|gif|jpg|jpeg|bmp|ico|ogm|ogg|mkv|mpg|mpeg|mp4|avi|asf|flv|wmv|qt))
            | (?P<word>  \b(?:[A-Z][a-z]+){2,}\b)
            | (?P<url>   (http|https|ftp|mailto)\:[^\s'\"]+\S)
            | (?P<email> [-\w._+]+\@[\w.-]+)

            # Lists, divs, spans
            | (?P<li>    ^\s+[\*\#]\s+)
            | (?P<pre>   \{\{\{|\s*\}\}\})
            | (?P<inl>   \{\{([^\s\|]+)(?:\s*\|\s*([^\]]+)|)\}\})

            # Tables
            | (?P<tr>    ^\s*\|\|(=|)\s*)
            | (?P<tre>   \s*\|\|(=|)\s*$)
            | (?P<td>    \s*\|\|(=|)\s*)

            # TODO: highlight search words (look at referrer)
          )""", re.VERBOSE)
        pre_re = re.compile("""(?:
              (?P<pre>\s*\}\}\})
            | (?P<ent>[<>&])"
            )""", re.VERBOSE)
        blank_re = re.compile(r"^\s*$")
        indent_re = re.compile(r"^\s*")
        tr_re = re.compile(r"^\s*\|\|")
        eol_re = re.compile(r"\r?\n")
        # For each line, we scan through looking for magic strings, outputting verbatim any intervening text
        #3.0: for self.line in eol_re.split(str(self.raw.expandtabs(), 'utf-8')):
        for self.line in eol_re.split(str(self.raw.expandtabs())):
            # Skip pragmas
            if self.in_header:
                if self.line.startswith('#'):
                    continue
                self.in_header = False

            if self.in_pre:
                print(re.sub(pre_re, self.replace, self.line))
            else:
                if self.in_table and not tr_re.match(self.line):
                    self.in_table = False
                    print('</tbody></table><p>')

                if blank_re.match(self.line):
                    print('</p><p>')
                else:
                    indent = indent_re.match(self.line)
                    #3.0: print(self._indent_to(len(indent.group(0))), end=' ')
                    print(self._indent_to(len(indent.group(0))))
                    print(re.sub(scan_re, self.replace, self.line))

        if self.in_pre: print('</pre>')
        if self.in_table: print('</tbody></table><p>')
        print(self._undent())
        print('</p></div>')

class Page:
    def __init__(self, page_name):
        self.page_name = page_name
        self.msg_text = ''
        self.msg_type = 'error'

    def split_title(self):
        # look for the end of words and the start of a new word and insert a space there
        return re.sub('([a-z])([A-Z])', r'\1 \2', self.page_name)

    def _filename(self):
        return os.path.join(data_dir, self.page_name)

    def _tmp_filename(self):
        return os.path.join(data_dir, ('#' + self.page_name.replace('/','_') + '.' + str(os.getpid()) + '#'))

    def exists(self):
        try:
            os.stat(self._filename())
            return True
        except OSError as err:
            if err.errno == errno.ENOENT:
                return False
            raise err

    def get_raw_body(self, default=None):
        try:
            return open(self._filename(), 'rb').read()
        except IOError as err:
            if err.errno == errno.ENOENT:
                if default is None:
                    default = '//[[%s|Describe %s|action=edit]]//' % (self.page_name, self.page_name)
                return default
            if err.errno == errno.EISDIR:
                return self.format_dir()
            raise err

    def format_dir(self):
        out = '== '
        pathname = ''
        for dirname in self.page_name.split('/'):
            pathname = (pathname + '/' + dirname) if pathname else dirname
            out += '[[' + pathname + '|' + dirname + ']]/'
        out += ' ==\n'
 
        for filename in page_list(self._filename(), file_re):
            if img_re.match(filename):
                if image_maxwidth:
                    maxwidth_arg = ' | maxwidth=' + str(image_maxwidth)
                out += '{{' + self.page_name + '/' + filename + ' | ' + humanlink(filename) + maxwidth_arg + ' | class=thumbleft}}\n'
            else:
                out += ' * [[' + self.page_name + '/' + filename + ']]\n'
        return out

    def pragmas(self):
        if not '_pragmas' in self.__dict__:
            self._pragmas = {}
            try:
                file = open(self._filename(), 'rt')
                attr_re = re.compile(r"^#(\S*)(.*)$")
                for line in file:
                    m = attr_re.match(line)
                    if not m:
                        break
                    self._pragmas[m.group(1)] = m.group(2).strip()
                    #print "bernie: pragmas[" + m.group(1) + "] = " + m.group(2) + "<br>\n"
            except IOError as err:
                if err.errno != errno.ENOENT and err.errno != errno.EISDIR:
                    raise er
        return self._pragmas

    def pragma(self, name, default):
        return self.pragmas().get(name, default)

    def can(self, action, default=True):
        acl = None
        try:
            #acl SomeUser:read,write All:read
            acl = self.pragma("acl", None)
            for rule in acl.split():
                (user, perms) = rule.split(':')
                if user == remote_user() or user == "All":
                    return action in perms.split(',')
            return False
        except Exception:
            if acl:
                self.msg_text = 'Illegal acl line: ' + acl
        return default

    def can_write(self):
        return self.can("write", True)

    def can_read(self):
        return self.can("read", True)

    def send_naked(self):
        if self.can_read():
            WikiFormatter(self.get_raw_body()).print_html()
        else:
            send_guru("Read access denied by ACLs", "notice")

    def format(self):
        #css foo.css
        value = self.pragma("css", None)
        if value:
            global link_urls
            link_urls += [ [ "stylesheet", value ] ]

        send_title(self.page_name, self.split_title(),
            msg_text=self.msg_text, msg_type=self.msg_type, writable=self.can_write())
        self.send_naked()
        send_footer(self._last_modified())

    def _last_modified(self):
        try:
            from time import localtime, strftime
            modtime = localtime(os.stat(self._filename())[stat.ST_MTIME])
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise err
            return None
        return strftime(datetime_fmt, modtime)

    def send_editor(self, preview=None):
        send_title(None, 'Edit ' + self.split_title(), msg_text=self.msg_text, msg_type=self.msg_type)
        if not self.can_write():
            send_guru("Write access denied by ACLs", "error")
            return

        filename = ''
        if 'file' in form:
            filename = form['file'].value

        print(('<p><b>Editing ' + self.page_name
            + ' for ' + cgi.escape(remote_user())
            + ' from ' + cgi.escape(get_hostname(remote_host()))
            + '</b></p>'))
        print('<div class="editor"><form name="editform" method="post" enctype="multipart/form-data" action="%s">' % relative_url(self.page_name))
        print('<input type="hidden" name="edit" value="%s">' % (self.page_name))
        print('<input type="input" id="editor" name="changelog" value="Edit page %s" accesskey="c" /><br />' % (self.page_name))
        print('<textarea wrap="off" spellcheck="true" id="editor" name="savetext" rows="17" cols="100" accesskey="e">%s</textarea>' \
            % cgi.escape(preview or self.get_raw_body(default='')))
        print('<label for="file" accesskey="u">Or Upload a file:</label> <input type="file" name="file" value="%s" />' % filename)
        print("""
            <br />
            <input type="submit" name="save" value="Save" accesskey="s">
            <input type="submit" name="preview" value="Preview" accesskey="p" />
            <input type="reset" value="Reset" />
            <input type="submit" name="cancel" value="Cancel" />
            <br />
            </form></div>
            <script language="javascript">
            <!--
            document.editform.savetext.focus()
            //-->
            </script>
            """)
        print("<p>" + link_tag('EditingTips') + "</p>")
        if preview:
            print("<div class='preview'>")
            WikiFormatter(preview).print_html()
            print("</div>")
        send_footer()

    def send_raw(self, mimetype='text/plain', args=[]):
        if not self.can_read():
            send_title(None, msg_text='Read access denied by ACLs', msg_type='notice')
            return

        if 'maxwidth' in args:
            import subprocess
            emit_header(mimetype)
            sys.stdout.flush()
            subprocess.check_call(['gm', 'convert', self._filename(),
                '-scale', args['maxwidth'].value + ' >', '-'])
        else:
            body = self.get_raw_body()
            emit_header(mimetype)
            print(body)

    def _write_file(self, data):
        tmp_filename = self._tmp_filename()
        open(tmp_filename, 'wb').write(data)
        name = self._filename()
        if os.name == 'nt':
            # Bad Bill!  POSIX rename ought to replace. :-(
            try:
                os.remove(name)
            except OSError as err:
                if err.errno != errno.ENOENT: raise err
        os.rename(tmp_filename, name)

    def save(self, newdata, changelog):
        if not self.can_write():
            self.msg_text = 'Write access denied by ACLs'
            self.msg_type = 'error'
            return

        self._write_file(newdata)
        rc = 0
        if post_edit_hook:
            import subprocess
            cmd = [ post_edit_hook, data_dir + '/' + self.page_name, remote_user(), remote_host(), changelog]
            child = subprocess.Popen(cmd, stdout=subprocess.PIPE, close_fds=True)
            output = child.stdout.read()
            rc = child.wait()
        if rc:
            self.msg_text += "Post-editing hook returned %d. Command was:\n'%s'\n" % (rc, "' '".join(cmd))
            if output:
                self.msg_text += 'Output follows:\n' + output
        else:
            self.msg_text = 'Thank you for your contribution.  Your attention to detail is appreciated.'
            self.msg_type = 'success'

try:
    exec(open("geekigeeki.conf.py").read())
    form = cgi.FieldStorage()
    action = form.getvalue('a', 'get')
    handler = globals().get('handle_' + action)
    if handler:
        handler(query_string(), form)
    else:
        send_httperror("403 Forbidden", query_string())

except Exception:
    import traceback
    msg_text = traceback.format_exc()
    if title_done:
        send_guru(msg_text, "error")
    else:
        send_title(None, msg_text=msg_text)
    send_footer()

sys.stdout.flush()
